/**
 * @file algorithm.hpp
 * @brief Modified multi-pass algorithm source code from seminars
 * for Rectilinear Texture Warping for Fast Adaptive Shadow Mapping.
 * 
 */
#pragma once

#include "common.h"
#include <map>

#define RENDERPASS_CALLBACK(callback) std::bind(&callback, this, std::placeholders::_1)

struct Algorithm;

using OptionsMap = std::map<std::string, bool>;
using VSInvocationCounter = Tools::GPUVSInvocationQuery;
using FSInvocationCounter = Tools::GPUFSInvocationQuery;
using CSInvocationCounter = Tools::GPUCSInvocationQuery;

/**
 * @brief Render pass to be included in a multi-pass algorithm.
 * 
 */
struct RenderPass {
    using renderPassFn = std::function<void(void)>; // Render function declaration

    /** @brief Renderpass name. */
    std::string          name;
    /** @brief Renderpass controller (optional). */
    const bool*          controller = nullptr;
    /** @brief Vertex shader file name. */
    std::string          vsFile;
    /** @brief Fragment shader file name. */
    std::string          fsFile;
    /** @brief Compute shader file name. */
    std::string          csFile;
    /** @brief Shader program ID. */
    GLuint               program = 0;
    /** @brief Render function. */
    renderPassFn         render;
    /** @brief GPU timer. */
    Tools::GPUTimer      timer;
    /** @brief GPU vertex shader execution counter. */
    VSInvocationCounter  vertexShaderCounter;
    /** @brief GPU fragment shader execution counter. */
    FSInvocationCounter  fragmentShaderCounter;
    /** @brief GPU compute shader execution counter. */
    CSInvocationCounter  computeShaderCounter;

    // Constructors
    /**
     * @brief Construct a new Render Pass.
     * 
     * @param _name name of the render pass
     * @param _shaderFile name of shader files (vs and fs, same)
     * @param _render render function to be called on run()
     */
    RenderPass(const std::string& _name, const char* _shaderFile, renderPassFn _render)
        : name(_name), controller(nullptr), vsFile(_shaderFile ? _shaderFile : ""), fsFile(_shaderFile ? _shaderFile : ""), render(_render) {}
    /**
     * @brief Construct a new Render Pass.
     * 
     * @param _name name of the render pass
     * @param _render render function to be called on run()
     */
    RenderPass(const std::string& _name, renderPassFn _render)
        : name(_name), controller(nullptr), vsFile(""), fsFile(""), render(_render) {}
    /**
     * @brief Construct a new Render Pass.
     * 
     * @param _name name of the render pass
     * @param _shaderFile  name of shader files (vs and fs, same)
     * @param _controller boolean condition
     * @param _render render function to be called on run()
     */
    RenderPass(const std::string& _name, const char* _shaderFile, const bool& _controller, renderPassFn _render)
        : name(_name), controller(&_controller), vsFile(_shaderFile ? _shaderFile : ""), fsFile(_shaderFile ? _shaderFile : ""), render(_render) {}
    /**
     * @brief Construct a new Render Pass.
     * 
     * @param _name name of the render pass
     * @param _controller boolean condition
     * @param _render render function to be called on run()
     */
    RenderPass(const std::string& _name, const bool& _controller, renderPassFn _render)
        : name(_name), controller(&_controller), vsFile(""), fsFile(""), render(_render) {}
    /**
     * @brief Construct a new Render Pass.
     * 
     * @param _name name of the render pass
     * @param _vsFile name of vertex shader file
     * @param _fsFile name of fragment shader file
     * @param _csFile name of compute shader file
     * @param _render render function to be called on run()
     */
    RenderPass(const std::string& _name, const char* _vsFile, const char* _fsFile, const char* _csFile, renderPassFn _render)
        : name(_name), controller(nullptr), vsFile(_vsFile ? _vsFile : ""), fsFile(_fsFile ? _fsFile : ""), csFile(_csFile ? _csFile : ""), render(_render) {}

    /**
     * @brief Run the assigned render function. Performance measurements optional.
     * 
     * @param algorithm 
     * @param trackStats 
     * @param forceSync 
     */
    void run(Algorithm& algorithm, bool trackStats = false, bool forceSync = false) {
        if (controller && (controller[0] == false)) return;

        if (trackStats) {
            vertexShaderCounter.start();
            fragmentShaderCounter.start();
            computeShaderCounter.start();
        }
        timer.start(forceSync);
            glUseProgram(program);
            render();
        timer.stop();
        if (trackStats) {
            computeShaderCounter.stop();
            fragmentShaderCounter.stop();
            vertexShaderCounter.stop();
        }
    }

    /**
     * @brief Reset timer.
     * 
     */
    void reset() {
        timer.reset();
    }

    /**
     * @brief Compile shaders in the pass. Tries vs and fs, cs on fail.
     * 
     * @param options possible defines addition
     * @return true success
     * @return false no shaders compiled
     */
    bool compile(const OptionsMap& options) {
        reset();
        if (vsFile.empty() && fsFile.empty() && csFile.empty())
            return true;

        // Create list of GLSL defines for enabled options
        std::string preprocessor;
        for (auto& option : options) {
            if (option.second) {
                std::string defineName = option.first;
                std::replace(defineName.begin(), defineName.end(), ' ', '_');
                preprocessor += "#define " + defineName + "\n";
            }
        }

        // Try to compile VS+FS at first...
        bool result = Tools::Shader::CreateShaderProgramFromFile(program, (vsFile + ".vert").c_str(), nullptr, nullptr, nullptr, (fsFile + ".frag").c_str(), preprocessor.empty() ? nullptr : preprocessor.c_str());
        // if fails do CS
        if (!result)
            result = Tools::Shader::CreateComputeShaderProgramFromFile(program, (csFile + ".txt").c_str());

        // Explicit sync. for better time measurement
        glFinish();
        return result;
    }

    /**
     * @brief Pass allowed to run.
     * 
     * @return true controller true
     * @return false controller exists and false
     */
    bool isEnabled() const {
        return !controller || (controller[0]);
    }
}; // end of struct RenderPass

/**
 * @brief Multipass algorithm definition.
 * 
 */
struct Algorithm {
    /** @brief Algorithm name. */
    std::string             name = "Algorithm";
    /** @brief Flag for lazy initialization of shaders. */
    bool                    initialized = false;
    /** @brief Flag if debug method will be called. */
    bool                    showDebug = false;
    /** @brief GPU timer for the complete algorithm. */
    Tools::GPUTimer         timer;
    /** @brief Render passes of the algorithm. */
    std::vector<RenderPass> renderPasses;
    /** @brief Algorithm options. */
    OptionsMap              options;

    // Constructor
    /**
     * @brief Construct a new Algorithm.
     * 
     * @param _name algorithm name
     */
    Algorithm(const std::string& _name) : name(_name) {}

    /**
     * @brief Initializes the algorithm
     * 
     */
    virtual void setup() {

    };
    
    /**
     * @brief Called at the end of run().
     * 
     */
    virtual void postRender() {}

    /**
     * @brief Window resized callback.
     * 
     * @param resolution new resolution
     */
    virtual void windowResized(const glm::ivec2& resolution) {};

    /**
     * @brief Executes the algorithm (calls run() on all included passes).
     * 
     * @param trackStats bool - produce pass statistics?
     * @param forceSync 
     */
    virtual void run(bool trackStats = false, bool forceSync = false) {
        // Lazy initialization of the algorithm
        if (!initialized) initialized = reset(true);

        timer.start(forceSync);
        for (auto& renderPass : renderPasses)
            renderPass.run(*this, trackStats, forceSync);
        timer.stop();

        // Render/print debug information
        postRender();
        
    }

    /**
     * @brief Resets algorithm - resets all performance timers/counters and restarts all renderpasses
     * (and optionally rebuilds shaders of all renderpasses).
     * 
     * @param resetShaders bool
     * @return true not compiling or recompile success
     * @return false recompile fail
     */
    virtual bool reset(bool resetShaders) {
        timer.reset();
        bool result = true;
        if (resetShaders) {
            return compile();
        } 
        for (auto& renderPass : renderPasses)
            renderPass.reset();
        return true;
    }

    /**
     * @brief Rebuilds algorithm shaders (rebuild shaders of all renderpasses).
     * 
     * @return true recompile success
     * @return false recompile fail
     */
    virtual bool compile() {
        for (auto& renderPass : renderPasses)
            if (!renderPass.compile(options)) return false;
        return true;
    }

    /**
     * @brief Displays window with GUI for the algorithm
     * 
     * @param position 
     * @param width 
     */
    virtual void gui(const glm::ivec2& position, int width) {
        // Count enabled renderpasses
        int numRenderPasses = 0;
        for (auto& renderPass : renderPasses) {
            if (renderPass.isEnabled()) numRenderPasses++;
        }

        int numControls = 3 * numRenderPasses + options.size();
        int const rowHeight = 10;
        int const rowSpace = 8;
        size_t const height = (170 + rowHeight * numControls + rowSpace * (numControls - 1));
        ImGui::Begin(name.c_str());
        ImGui::SetWindowPos(ImVec2(position.x, position.y) * IMGUI_RESIZE_FACTOR, ImGuiCond_Once);
        ImGui::SetWindowSize(ImVec2(width, height) * IMGUI_RESIZE_FACTOR, ImGuiCond_Always);
        // Add checkboxes for algorithm options
        ImGui::Text("OPTIONS");
        for (auto& option : options) {
            if (ImGui::Checkbox(option.first.c_str(), &option.second))
                reset(true);
        }
        ImGui::Separator();
        // Timers
        ImGui::Text("TIMERS");
        auto const totalTime = timer.getAverage();
        ImGui::Text("Frame time: %u (%u)", totalTime, timer.getCounter());
        for (auto& renderPass : renderPasses) {
            if (renderPass.isEnabled()) {
                auto const renderPassTime = renderPass.timer.getAverage();
                ImGui::Text(" * %s: %u (%.2f%%)", renderPass.name.c_str(), renderPassTime, float(renderPassTime) / totalTime * 100.0f);
            }
        }
        // Invocation counters
        ImGui::Separator();
        ImGui::Text("VERTEX SHADERS");
        for (auto& renderPass : renderPasses) {
            if (renderPass.isEnabled())
                ImGui::Text(" * %s: %u", renderPass.name.c_str(), renderPass.vertexShaderCounter.get());
        }
        ImGui::Separator();
        ImGui::Text("FRAGMENT SHADERS");
        for (auto& renderPass : renderPasses) {
            if (renderPass.isEnabled())
                ImGui::Text(" * %s: %u", renderPass.name.c_str(), renderPass.fragmentShaderCounter.get());
        }
        ImGui::Separator();
        ImGui::Text("COMPUTE SHADERS");
        for (auto& renderPass : renderPasses) {
            if (renderPass.isEnabled())
                ImGui::Text(" * %s: %u", renderPass.name.c_str(), renderPass.computeShaderCounter.get());
        }
        ImGui::End();
    }

}; // end of struct Algorithm


