/**
 * @file controls.hpp
 * @brief Modified controls source code from seminars
 * for Rectilinear Texture Warping for Fast Adaptive Shadow Mapping.
 * 
 */

const char* help_message =
"Rectilinear Texture Warping for Fast Adaptive Shadow Mapping\n\
-------------------------------------------------------------------------------\n\
CONTROLS:\n\
   [z/Z]   ... move scene along z-axis\n\
   [s]     ... show/hide computed textures\n\
   [t]     ... show/hide and compute/sleep for passes stats\n\
   [a/A]   ... algorithm choice\n\
   [f/F]   ... float visualization scale\n\
   [y/Y]   ... z offset for shadow mapping\n\
   [d/D]   ... change depth map resolution\n\
   [c]     ... compile shaders\n\
   [mouse] ... scene rotation (left button)\n\
-------------------------------------------------------------------------------";

// IMPLEMENTATION______________________________________________________________

/**
 * @brief Update light matrix (serves as directional light control, looking at the world origin from g_LightPostiion).
 * 
 */
void updateLightViewMatrix() {
    // Compute light view matrix
    g_LightViewMatrix = glm::lookAt(g_LightPosition, glm::vec3(0.0f, 0.0f, 0.0f), glm::normalize(glm::vec3(-g_LightPosition.y, g_LightPosition.x, 0.0f)));
}


/**
 * @brief GUI rendering.
 * 
 * @return int 
 */
int showGUI() {
    int const menuHeight = 210 + 190;

    ImGui::Begin("Render");
    ImGui::SetWindowSize(ImVec2(220, menuHeight) * IMGUI_RESIZE_FACTOR, ImGuiCond_Always);
        ImGui::Checkbox("show texture", &g_ShowTextures);
        ImGui::Checkbox("show alg stats", &g_ShowPassesStats);
    ImGui::Separator();
    ImGui::Text(" FRAMEBUFFER");
        ImGui::SetNextItemWidth(120 * IMGUI_RESIZE_FACTOR);
        int resolution = int(glm::round(glm::log2(g_Resolution / 32.0f)));
        if (ImGui::Combo("resolution", &resolution, " 32x32\0 64x64\0 128x128 \0 256x256\0 512x512 \0 1024x1024\0 2048x2048\0")) {
            g_Resolution = 32 << resolution;
            createVirtualFramebuffers();
        }
    ImGui::Separator();
    ImGui::Text(" LIGHT");
        ImGui::SetNextItemWidth(190 * IMGUI_RESIZE_FACTOR);
        ImGui::SliderFloat("x", &g_LightPosition.x, -30.0f, 30.0f, "%.3f");
        ImGui::SetNextItemWidth(190 * IMGUI_RESIZE_FACTOR);
        ImGui::SliderFloat("y", &g_LightPosition.y, -30.0f, 30.0f, "%.3f");
        ImGui::SetNextItemWidth(190 * IMGUI_RESIZE_FACTOR);
        ImGui::SliderFloat("z", &g_LightPosition.z, -30.0f, 30.0f, "%.3f");
    ImGui::Separator();
    ImGui::Text(" APP CONTROL");
    ImGui::SetNextItemWidth(120 * IMGUI_RESIZE_FACTOR);
        int algChoice = g_Algorithm;
        if (ImGui::Combo("algorithm", &algChoice, " classic\0 forward\0 backwards\0")) {
            g_Algorithm = algChoice;
        }
        if (g_Algorithm > Classic)
        {
            ImGui::SetNextItemWidth(120 * IMGUI_RESIZE_FACTOR);
            ImGui::SliderFloat("vis. scale", &g_FloatVisScale, 1.f, 20.0f, "%0.1f");
            ImGui::SetNextItemWidth(120 * IMGUI_RESIZE_FACTOR);
            unsigned blurMin = 2;
            unsigned blurMax = 6;
            ImGui::SliderScalar("blur power", ImGuiDataType_U32, &g_BlurPower, &blurMin, &blurMax, "%u");
        }
    ImGui::Separator();
    ImGui::Text(" SHADOWMAP");
        ImGui::SetNextItemWidth(120 * IMGUI_RESIZE_FACTOR);
        ImGui::SliderFloat("offset", &g_zOffset, 0.f, 0.025f, "%.8f");

    ImGui::End();

    // additional stats on multi-pass rendering.
    if (g_ShowPassesStats)
    {
        if (g_Algorithm == Classic)
            g_ClassicShadowMap.gui(glm::ivec2(235, 10), 260);
        else if (g_Algorithm == Backwards)
            g_RTWBackwards.gui(glm::ivec2(235, 10), 260);
        else if (g_Algorithm == Forward)
            g_RTWForward.gui(glm::ivec2(235, 10), 260);
    }

    return menuHeight;
}


/**
 * @brief Reset algorithms.
 * 
 * @param resetShaders reset shaders too
 */
void resetAlgorithms(bool resetShaders = false) {
    g_ClassicShadowMap.reset(resetShaders);
    g_RTWBackwards.reset(resetShaders);
    g_RTWForward.reset(resetShaders);
}


/**
 * @brief Compile shaders (aka reset algorithms).
 * 
 * @param clientData 
 */
void compileShaders(void* clientData) {
    resetAlgorithms(true);
}


/**
 * @brief Keyboard input control.
 * 
 * @param key val
 * @param action 
 * @param mods shift?
 */
void keyboardChanged(int key, int action, int mods) {
    switch (key) {
    case GLFW_KEY_S: g_ShowTextures = !g_ShowTextures; break;
    case GLFW_KEY_T: g_ShowPassesStats = !g_ShowPassesStats; break;
    case GLFW_KEY_D: g_Resolution = (mods == GLFW_MOD_SHIFT) ? ((g_Resolution == 32) ? 2048 : (g_Resolution >> 1)) : ((g_Resolution == 2048) ? 32 : (g_Resolution << 1));
        createVirtualFramebuffers();
        break;
    case GLFW_KEY_A: g_Algorithm = (mods == GLFW_MOD_SHIFT) ? ((g_Algorithm == 0) ? 2 : (g_Algorithm - 1)) : ((g_Algorithm == 2) ? 0 : g_Algorithm + 1); break;
    case GLFW_KEY_F: g_FloatVisScale = (mods == GLFW_MOD_SHIFT) ? ((g_FloatVisScale >= 1.5f) ? (g_FloatVisScale - 0.5f) : 1.f) : ((g_FloatVisScale <= 19.5f) ? (g_FloatVisScale + 0.5f) : 20.f); break;
    case GLFW_KEY_Y: g_zOffset = (mods == GLFW_MOD_SHIFT) ? ((g_zOffset >= 0.0005f) ? (g_zOffset - 0.0005f) : 0.0f) : ((g_zOffset <= 0.0245f) ? (g_zOffset + 0.0005f) : 0.025f); break;
    }
}


/**
 * @brief Controls main function.
 * 
 * @param argc 
 * @param argv 
 * @return int 
 */
int main(int argc, char* argv[]) {
    int OGL_CONFIGURATION[] = {
        GLFW_CONTEXT_VERSION_MAJOR,  4,
        GLFW_CONTEXT_VERSION_MINOR,  5,
        GLFW_OPENGL_FORWARD_COMPAT,  GL_FALSE,
        GLFW_OPENGL_DEBUG_CONTEXT,   GL_TRUE,
        GLFW_OPENGL_PROFILE,         GLFW_OPENGL_COMPAT_PROFILE, // GLFW_OPENGL_CORE_PROFILE
        PGR2_SHOW_MEMORY_STATISTICS, GL_TRUE, 
        0
    };

    printf("%s\n", help_message);

    return common_main(1200, 900, "Rectilinear Texture Warping for Fast Adaptive Shadow Mapping",
                       OGL_CONFIGURATION, // OGL configuration hints
                       init,              // Init GL callback function
                       nullptr,           // Release GL callback function
                       showGUI,           // Show GUI callback function
                       display,           // Display callback function
                       nullptr,           // Window resize callback function
                       keyboardChanged,   // Keyboard callback function
                       nullptr,           // Mouse button callback function
                       nullptr);          // Mouse motion callback function
}
