/**
 * @file shadow_mapping.cpp
 * @author hlavoter
 * @brief Rectilinear Texture Warping for Fast Adaptive Shadow Mapping implementation.
 * @date 2023/2024 winter semester for PGR2
 * 
 */

#include "algorithm.hpp"

// GLOBAL CONSTANTS____________________________________________________________
/**
 * @brief Diffuse texture's filepath to be applied on scene models.
 */
const char* TEXTURE_FILE_NAME = "../shared/textures/metal01.jpg";
/**
 * @brief Texture enum definition needed for the multi-pass shadowmap rendering algorithms.
 */
enum eTextureType { Diffuse = 0, Shadow, ShadowZ, PreImportance, PreImportanceZ, Importance, WarpX, WarpY, BlurWarpX, BlurWarpY, StretchWarpX, StretchWarpY, NumTextureTypes };
/**
 * @brief Frame buffer enum definition needed for the multi-pass shadowmap rendering algorithms.
 */
enum eFrameBufferType { ShadowFB = 0, FullRenderFB, ImportanceFB, NumFBTypes };
/**
 * @brief Multi-pass shadowmap rendering algorithm enum definition.
 */
enum eAlgorithmType { Classic = 0, Forward, Backwards, NumAlgorithmTypes };

// GLOBAL VARIABLES____________________________________________________________

// GUI control
/** Show/hide computed textures. */
bool      g_ShowTextures              = true;
/** Show and compute stats for individual passes. */
bool      g_ShowPassesStats           = true;
// shadows control
/** Z offset for rendering shadowmaps and minimize artefacts. */
float     g_zOffset                 = 0.005f;
/** Algorithm type, user-chosen. */
GLint     g_Algorithm              = Classic;
/** FBO/Shadow mapping size in pixels. */
GLint     g_Resolution                = 1024;
/** Float scale for uint < -- > float conversions. */
float     g_FloatConvScale           = 4000000.f;
/** Float scale for importance/warping maps visualisations. */
float     g_FloatVisScale             =  1.f;
/** How much the warp map gets blurred. */
GLuint    g_BlurPower                 =    3;

// Textures and framebuffers
/** Textures - color (diffuse), depth map, z-buffer, z-buffer for hw shadow map. */
GLuint    g_Textures[NumTextureTypes] =  {0};
/** Virtual frame-buffer FBO id. */
GLuint    g_Framebuffer[NumFBTypes]   =  {0};

// Transformations
/** Camera view transformation ~ scene transformation. */
glm::mat4& g_CameraViewMatrix       = Variables::Transform.ModelView;
/** Camera projection transformation ~ scene projection. */
glm::mat4& g_CameraProjectionMatrix = Variables::Transform.Projection;
/** Light orientation. */
glm::vec3  g_LightPosition          = glm::vec3(0.0f, 20.0f, 0.0f);
/** Light view transformation. */
glm::mat4  g_LightViewMatrix;
/** Light projection transformation. */
glm::mat4  g_LightProjectionMatrix  = glm::frustum(-1.0f, 1.0f,-1.0f, 1.0f, 1.0f, 1000.0f);

// Precomputed variables
/** Projection view model matrix, camera. */
glm::mat4 g_PVM;
/** Inverse to projection view model matrix, camera. */
glm::mat4 g_inversePMV;
/** Projection view model matrix, light. */
glm::mat4 g_lightPVM;
/** Inverse to projection view model matrix, light. */
glm::mat4 g_inverseLightPVM;
/** Position of the light. */
glm::vec4 g_light_position;
/** Window size in float. */
glm::vec2 g_window;
/** Previous window size - to resize textures if needed. */
glm::ivec2 g_prevWindow;
/** Compute shader workgroup size, depends on shadow map resolution. */
int g_workgroup;

// FORWARD DECLARATIONS________________________________________________________
void createVirtualFramebuffers();
void updateLightViewMatrix();
void checkFramebuffer(GLuint type, GLint res);

// IMPLEMENTATION______________________________________________________________

/**
 * @brief Classic conventional shadow mapping algorithm implementation - 2 passes.
 * 
 */
struct ClassicShadowmapAlgorithm : Algorithm {

    ClassicShadowmapAlgorithm() : Algorithm("ClassicShadowMap") {}

    void setup() override {
        showDebug = true;

        // first render in light POV
        renderPasses.emplace_back("Shadowmap generation", "depth_render", "depth_render", nullptr, [&]() -> void {
            glBindFramebuffer(GL_FRAMEBUFFER, g_Framebuffer[ShadowFB]);
            glViewport(0, 0, g_Resolution, g_Resolution);
            glClearColor(1.f, 0.f, 0.f, 0.f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_lightPVM[0][0]);
            Tools::DrawScene();
            });

        // then use the rendered shadow map in full render from camera
        renderPasses.emplace_back("Full render", "classic_shadows_full_render", "classic_shadows_full_render", nullptr, [&]() -> void {
            glViewport(0, 0, Variables::WindowSize.x, Variables::WindowSize.y);
            glBindFramebuffer(GL_FRAMEBUFFER, 0);
            glClearColor(0.65f, 0.81f, 0.60f, 1.0f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glBindTextureUnit(0, g_Textures[Diffuse]);
            glBindTextureUnit(1, g_Textures[Shadow]);
            
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_CameraViewMatrix[0][0]);
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_CameraProjectionMatrix[0][0]);
            glUniformMatrix4fv(2, 1, GL_FALSE, &g_LightViewMatrix[0][0]);
            glUniformMatrix4fv(3, 1, GL_FALSE, &g_LightProjectionMatrix[0][0]);
            
            glUniform4fv(4, 1, &g_light_position.x);
            glUniform1f(5, g_zOffset);  // offset for shadow detection - to combat self shadowing/surface acne

            Tools::DrawScene();
            });

    }

    void postRender() override {
        if (g_ShowTextures)
            Tools::Texture::Show2D(g_Textures[Shadow], Variables::WindowSize.x - 200, Variables::WindowSize.y - 200, 200, 200);
    }
};
ClassicShadowmapAlgorithm g_ClassicShadowMap;


/**
 * @brief Rectilinear Texture Warping shadowmap algorithm implementation - variant backwards.
 * 0th pass - clear.
 * 1st pass pre-importance: full shadowless render with camera PVM.
 * 2nd pass importance: project texels "back" into light space, save importance there.
 * 3rd pass warp I.: save maximum values into two 1D textures (as row and column).
 * 4th pass warp II.: blur 1D textures.
 * 5th pass warp III.: compute warp values, save to 1D textures.
 * 6th pass warped shadow map: finally create shadow map with shifted vertices.
 * 7th pass final composition: add shadows to the rendered image from 1st pass based on the warped shadowmap.
 */
struct RTWBackwardsAlgorithm : Algorithm {

    RTWBackwardsAlgorithm() : Algorithm("BackwardsRTWShadowMap") {}

    void setup() override {
        showDebug = true;

        // clear textures that do not get overwritten in their entirity in the following passes
        renderPasses.emplace_back("Clear Resources", nullptr, [&]() -> void {
            uint32_t zero = 0;
            glClearTexImage(g_Textures[Importance], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero); // clearing the importance texture
            glClearTexImage(g_Textures[WarpX], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero);
            glClearTexImage(g_Textures[WarpY], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero);
        });

        /**
         * 1. IMPORTANCE.
         */
        // pre importance pass - backwards algorithm version renders from camera perspective fully as the first step
        renderPasses.emplace_back("Pre importance pass", "no_shadows_full_render", "no_shadows_full_render", nullptr, [&]() -> void {
            // backwards render does full render (without shadows) from the camera's POV
            // renders full frame with depth.
            glViewport(0, 0, Variables::WindowSize.x, Variables::WindowSize.y);
            glBindFramebuffer(GL_FRAMEBUFFER, g_Framebuffer[FullRenderFB]);
            glClearColor(0.65f, 0.81f, 0.60f, 1.0f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glBindTextureUnit(0, g_Textures[Diffuse]); 
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_CameraViewMatrix[0][0]);
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_CameraProjectionMatrix[0][0]);
            glUniform4fv(2, 1, &g_light_position.x);

            Tools::DrawScene();
        });

        // importance pass - now the depth component of the previous render gets analyzed and the values get projected into light space
        renderPasses.emplace_back("Importance pass", "copy_shader", "importance_backwards_pass", nullptr, [&]() -> void {
            glBindFramebuffer(GL_FRAMEBUFFER, 0);
            glDisable(GL_DEPTH_TEST);
            glViewport(0, 0, Variables::WindowSize.x, Variables::WindowSize.y); // process whole rendered depth image
            glClearColor(0.f, 0.f, 0.f, 0.f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            // shader uses atomic operations to write the transformed values into the importance texture
            // so the format needs to be integer one channel
            glBindImageTexture(1, g_Textures[Importance], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);
            glBindTextureUnit(0, g_Textures[PreImportanceZ]);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_inversePMV[0][0]);  // camera space to world space transform
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_lightPVM[0][0]);  // world space to light space transform
            glUniform2fv(2, 1, &(g_window.x));
            glUniform1f(3, g_Resolution);
            glUniform1f(4, g_FloatConvScale);
            // works with previously rendered depth map, so there's no need to render scene
            glDrawArrays(GL_TRIANGLES, 0, 3);
            glEnable(GL_DEPTH_TEST);
        });

        /**
         * 2. WARP PASSES - into 1D textures, blurring, stretching.
         */
        // warping I. - importance function into column/row summation
        renderPasses.emplace_back("Warp pass I.", nullptr, nullptr, "warp1_pass", [&]() -> void {
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
            glBindImageTexture(0, g_Textures[Importance], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[WarpX], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);  // read/write because of atomics
            glBindImageTexture(2, g_Textures[WarpY], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);
            // this pass uses imageAtomicMax to store maximum importance from previously rendered
            // importance map to the two 1D textures.
            glDispatchCompute(g_workgroup, g_workgroup, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        // warping II. - blur
        renderPasses.emplace_back("Warp pass II.", nullptr, nullptr, "warp2_pass", [&]() -> void {
            glUniform1ui(0, g_BlurPower);
            glBindImageTexture(0, g_Textures[WarpX], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[BlurWarpX], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32UI);
            // previously computed 1D textures get blurred in this step. There's no need for atomic operations
            // as every thread writes into its assigned texel while reading from multiple in the read only 1D texture from previous pass.
            glDispatchCompute(g_workgroup, 1, 1);
            glBindImageTexture(0, g_Textures[WarpY], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[BlurWarpY], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32UI);
            glDispatchCompute(g_workgroup, 1, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        // warping III. - stretch
        renderPasses.emplace_back("Warp pass III.", nullptr, nullptr, "warp3_pass", [&]() -> void {
            glUniform1f(0, g_FloatConvScale);
            glBindImageTexture(0, g_Textures[BlurWarpX], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[StretchWarpX], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RG32F);
            // in this step, warp gets computed for every texel. Again, no atomic operations are needed.
            glDispatchCompute(g_workgroup, 1, 1);
            glBindImageTexture(0, g_Textures[BlurWarpY], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[StretchWarpY], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RG32F);
            glDispatchCompute(g_workgroup, 1, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        /**
         * 3. WARPED SHADOW MAP GENERATION.
         */
        // in this step, shadow map is generated, but in the vertex shader, all vertices are translated based on warp
        // values from the 1D warp textures to generate more fragments in the parts of the shadow map texture
        // with the high importance values computed. This will ensure higher resolution in those parts and thus
        // higher resolution shadows will be placed into the scene in desired places (close to camera in this case)
        renderPasses.emplace_back("Warped map", "warped_shadow_gen_pass", "warped_shadow_gen_pass", nullptr, [&]() -> void {
            glBindFramebuffer(GL_FRAMEBUFFER, g_Framebuffer[ShadowFB]);
            glViewport(0, 0, g_Resolution, g_Resolution);
            glClearColor(1.f, 0.f, 0.f, 0.f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);     
            glBindTexture(GL_TEXTURE_2D, g_Textures[StretchWarpX]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glBindTexture(GL_TEXTURE_2D, g_Textures[StretchWarpY]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glBindTextureUnit(0, g_Textures[StretchWarpX]);
            glBindTextureUnit(1, g_Textures[StretchWarpY]);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_lightPVM[0][0]);
            glUniform1f(1, (float)g_Resolution);
            Tools::DrawScene();
        });

        /**
         * 4. FINAL RENDER COMPOSITION.
         */
        // previously rendered image from the first render pass gets now used and only shadows are composited to it in this step
        // from the warped shadow map (shifted uv coordinates are used based on the values from the two 1D warp textures)
        renderPasses.emplace_back("Final composition", "copy_shader", "shadow_render_backwards", nullptr, [&]() -> void {
            glViewport(0, 0, Variables::WindowSize.x, Variables::WindowSize.y);  
            glBindFramebuffer(GL_FRAMEBUFFER, 0);
            glClearColor(0.65f, 0.81f, 0.60f, 1.0f);
            glClear(GL_COLOR_BUFFER_BIT);
            glBindTextureUnit(0, g_Textures[PreImportance]);
            glBindTextureUnit(1, g_Textures[PreImportanceZ]);
            glBindTextureUnit(2, g_Textures[StretchWarpX]);
            glBindTextureUnit(3, g_Textures[StretchWarpY]);
            glBindTextureUnit(4, g_Textures[Shadow]);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_inversePMV[0][0]);
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_lightPVM[0][0]);
            glUniform1f(2, g_zOffset);
            glUniform2fv(3, 1, &(g_window.x));
            glDisable(GL_DEPTH_TEST);
            // no need to render the scene, it was already done in the beginning
            glDrawArrays(GL_TRIANGLES, 0, 3);
            glEnable(GL_DEPTH_TEST);
        });

    }

    void postRender() override {
        if (g_ShowTextures)
        {
            Tools::Texture::Show2D(g_Textures[Shadow], Variables::WindowSize.x - 200, Variables::WindowSize.y - 200, 200, 200);
            Tools::Texture::Show2D(g_Textures[Importance], Variables::WindowSize.x - 200, Variables::WindowSize.y - 400, 200, 200, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[WarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 450, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[WarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 500, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[BlurWarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 550, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[BlurWarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 600, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[StretchWarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 650, 200, 50);
            Tools::Texture::Show2D(g_Textures[StretchWarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 700, 200, 50);
        }
    }
};
RTWBackwardsAlgorithm g_RTWBackwards;


/**
 * @brief Rectilinear Texture Warping shadowmap algorithm implementation - variant forward.
 * 0th pass - clear.
 * 1st pass importance: render depth map from light and look "forward" to compute importance values.
 * 2nd pass warp I.: save maximum values into two 1D textures (as row and column).
 * 3rd pass warp II.: blur 1D textures.
 * 4th pass warp III.: compute warp values, save to 1D textures.
 * 5th pass warped shadow map: finally create shadow map with shifted vertices.
 * 6th pass final render: render full image with shadows based on the warped shadow map. 
 */
struct RTWForwardAlgorithm : Algorithm {

    RTWForwardAlgorithm() : Algorithm("ForwardRTWShadowMap") {}

    void setup() override {
        showDebug = true;

        renderPasses.emplace_back("Clear Resources", nullptr, [&]() -> void {
            uint32_t zero = 0;
            glClearTexImage(g_Textures[Importance], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero); // clearing the importance texture
            glClearTexImage(g_Textures[WarpX], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero);
            glClearTexImage(g_Textures[WarpY], 0, GL_RED_INTEGER, GL_UNSIGNED_INT, &zero);
        });

        /**
         * 1. IMPORTANCE.
         */
         // pre importance pass - backwards algorithm version renders from camera perspective fully as the first step
        renderPasses.emplace_back("Pre importance pass", "depth_render", "importance_forward_pass", nullptr, [&]() -> void {
            glBindFramebuffer(GL_FRAMEBUFFER, g_Framebuffer[ShadowFB]);
            glViewport(0, 0, g_Resolution, g_Resolution);
            glClearColor(1.f, 0.f, 0.f, 0.f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glBindImageTexture(0, g_Textures[Importance], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_lightPVM[0][0]);
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_inverseLightPVM[0][0]);
            glUniformMatrix4fv(2, 1, GL_FALSE, &g_PVM[0][0]);
            glUniform2fv(3, 1, &(g_window.x));
            glUniform1f(4, g_Resolution);
            glUniform1f(5, g_FloatConvScale);
            Tools::DrawScene();
        });

        /**
         * 2. WARP PASSES - into 1D textures, blurring, stretching.
         */
         // warping I. - importance function into column/row summation
        renderPasses.emplace_back("Warp pass I.", nullptr, nullptr, "warp1_pass", [&]() -> void {
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
            glBindImageTexture(0, g_Textures[Importance], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[WarpX], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);
            glBindImageTexture(2, g_Textures[WarpY], 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI);
            glDispatchCompute(g_workgroup, g_workgroup, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        // warping II. - blur
        renderPasses.emplace_back("Warp pass II.", nullptr, nullptr, "warp2_pass", [&]() -> void {
            glUniform1ui(0, g_BlurPower);
            glBindImageTexture(0, g_Textures[WarpX], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[BlurWarpX], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32UI);
            glDispatchCompute(g_workgroup, 1, 1);
            glBindImageTexture(0, g_Textures[WarpY], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[BlurWarpY], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32UI);
            glDispatchCompute(g_workgroup, 1, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        // warping III. - stretch
        renderPasses.emplace_back("Warp pass III.", nullptr, nullptr, "warp3_pass", [&]() -> void {
            glUniform1f(0, g_FloatConvScale);
            glBindImageTexture(0, g_Textures[BlurWarpX], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[StretchWarpX], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RG32F);
            glDispatchCompute(g_workgroup, 1, 1);
            glBindImageTexture(0, g_Textures[BlurWarpY], 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32UI);
            glBindImageTexture(1, g_Textures[StretchWarpY], 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RG32F);
            glDispatchCompute(g_workgroup, 1, 1);
            glMemoryBarrier(GL_ALL_BARRIER_BITS);
        });

        /**
         * 3. WARPED SHADOW MAP GENERATION.
         */
        renderPasses.emplace_back("Warped map", "warped_shadow_gen_pass", "warped_shadow_gen_pass", nullptr, [&]() -> void {
            glBindFramebuffer(GL_FRAMEBUFFER, g_Framebuffer[ShadowFB]);
            glViewport(0, 0, g_Resolution, g_Resolution);
            glClearColor(1.f, 0.f, 0.f, 0.f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glBindTexture(GL_TEXTURE_2D, g_Textures[StretchWarpX]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glBindTexture(GL_TEXTURE_2D, g_Textures[StretchWarpY]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glBindTextureUnit(0, g_Textures[StretchWarpX]);
            glBindTextureUnit(1, g_Textures[StretchWarpY]);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_lightPVM[0][0]);
            glUniform1f(1, (float)g_Resolution);
            Tools::DrawScene();
        });

        /**
         * 4. FINAL RENDER.
         */
        renderPasses.emplace_back("Final render", "classic_shadows_full_render", "shadow_render_forward", nullptr, [&]() -> void {
            glViewport(0, 0, Variables::WindowSize.x, Variables::WindowSize.y);
            glBindFramebuffer(GL_FRAMEBUFFER, 0);
            glClearColor(0.65f, 0.81f, 0.60f, 1.0f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glBindTextureUnit(0, g_Textures[Diffuse]);
            glBindTextureUnit(1, g_Textures[Shadow]);
            glBindTextureUnit(2, g_Textures[StretchWarpX]);
            glBindTextureUnit(3, g_Textures[StretchWarpY]);
            glUniformMatrix4fv(0, 1, GL_FALSE, &g_CameraViewMatrix[0][0]);
            glUniformMatrix4fv(1, 1, GL_FALSE, &g_CameraProjectionMatrix[0][0]);
            glUniformMatrix4fv(2, 1, GL_FALSE, &g_LightViewMatrix[0][0]);
            glUniformMatrix4fv(3, 1, GL_FALSE, &g_LightProjectionMatrix[0][0]);
            glUniform1f(4, g_zOffset);
            glUniform4fv(5, 1, &g_light_position.x);
            Tools::DrawScene();
        });

    }

    void postRender() override {
        if (g_ShowTextures)
        {
            Tools::Texture::Show2D(g_Textures[Shadow], Variables::WindowSize.x - 200, Variables::WindowSize.y - 200, 200, 200);
            Tools::Texture::Show2D(g_Textures[Importance], Variables::WindowSize.x - 200, Variables::WindowSize.y - 400, 200, 200, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[WarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 450, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[WarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 500, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[BlurWarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 550, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[BlurWarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 600, 200, 50, (g_FloatVisScale / g_FloatConvScale));
            Tools::Texture::Show2D(g_Textures[StretchWarpX], Variables::WindowSize.x - 200, Variables::WindowSize.y - 650, 200, 50);
            Tools::Texture::Show2D(g_Textures[StretchWarpY], Variables::WindowSize.x - 200, Variables::WindowSize.y - 700, 200, 50);
        }
    }
};
RTWForwardAlgorithm g_RTWForward;

/**
 * @brief Display function, to be called for every frame.
 * Executes tun() on chosen algorithm.
 */
void display() {
    // Recreate FB & textures if needed
    if (Variables::WindowSize.x != g_prevWindow.x || Variables::WindowSize.y != g_prevWindow.y)
    {
        g_prevWindow = Variables::WindowSize;
        g_window = glm::vec2(Variables::WindowSize.x, Variables::WindowSize.y);
        createVirtualFramebuffers();
    }

    // Update camera & light transformations
    updateLightViewMatrix();

    g_PVM = g_CameraProjectionMatrix * g_CameraViewMatrix;
    g_inversePMV = glm::inverse(g_PVM);
    g_lightPVM = g_LightProjectionMatrix * g_LightViewMatrix;
    g_inverseLightPVM = glm::inverse(g_lightPVM);
    g_light_position = (g_CameraViewMatrix * glm::inverse(g_LightViewMatrix)) * glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (g_Algorithm == Classic)
        g_ClassicShadowMap.run(g_ShowPassesStats);
    else if (g_Algorithm == Backwards)
        g_RTWBackwards.run(g_ShowPassesStats);
    else if (g_Algorithm == Forward)
        g_RTWForward.run(g_ShowPassesStats);

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

/**
 * @brief (Re)creates resources for the algorithms: textures, framebuffers, depth maps..
 * 
 */
void createVirtualFramebuffers() {
    static GLint s_Resolution = 0;

    g_workgroup = (g_Resolution + 7) / 8;

    if (s_Resolution == g_Resolution)
        return;

    glDeleteTextures(NumTextureTypes - 1, &g_Textures[Shadow]);
    glm::ivec2 resolution(g_Resolution, g_Resolution);

    // shadow map texture framebuffer
    glDeleteFramebuffers(1, &g_Framebuffer[ShadowFB]);
    Tools::Texture::Create2DDepth(g_Textures[Shadow], GL_R32F, resolution); // depth texture
    Tools::Texture::Create2DDepth(g_Textures[ShadowZ], GL_DEPTH_COMPONENT32F, resolution);
    // frame buffer init
    glCreateFramebuffers(1, &g_Framebuffer[ShadowFB]);
    glNamedFramebufferTexture(g_Framebuffer[ShadowFB], GL_COLOR_ATTACHMENT0, g_Textures[Shadow], 0);
    glNamedFramebufferTexture(g_Framebuffer[ShadowFB], GL_DEPTH_ATTACHMENT, g_Textures[ShadowZ], 0);
    // check framebuffer status
    checkFramebuffer(ShadowFB, s_Resolution);

    // pre importance framebuffer - full buffer
    glDeleteFramebuffers(1, &g_Framebuffer[FullRenderFB]);
    Tools::Texture::Create2D(g_Textures[PreImportance], GL_RGB32F, Variables::WindowSize);
    Tools::Texture::Create2DDepth(g_Textures[PreImportanceZ], GL_DEPTH_COMPONENT32F, Variables::WindowSize);
    // frame buffer init
    glCreateFramebuffers(1, &g_Framebuffer[FullRenderFB]);
    glNamedFramebufferTexture(g_Framebuffer[FullRenderFB], GL_COLOR_ATTACHMENT0, g_Textures[PreImportance], 0);
    glNamedFramebufferTexture(g_Framebuffer[FullRenderFB], GL_DEPTH_ATTACHMENT, g_Textures[PreImportanceZ], 0);
    // check framebuffer status
    checkFramebuffer(FullRenderFB, s_Resolution);

    Tools::Texture::Create2DDepth(g_Textures[Importance], GL_R32UI, resolution);

    // warp map
    glm::ivec2 warpresolution(g_Resolution, 1);
    Tools::Texture::Create2D(g_Textures[WarpX], GL_R32UI, warpresolution);
    Tools::Texture::Create2D(g_Textures[WarpY], GL_R32UI, warpresolution);
    Tools::Texture::Create2D(g_Textures[BlurWarpX], GL_R32UI, warpresolution);
    Tools::Texture::Create2D(g_Textures[BlurWarpY], GL_R32UI, warpresolution);
    Tools::Texture::Create2D(g_Textures[StretchWarpX], GL_RG32F, warpresolution);
    Tools::Texture::Create2D(g_Textures[StretchWarpY], GL_RG32F, warpresolution);
}

/**
 * @brief Check frame buffer for errors during creation.
 * 
 * @param type id
 * @param res resolution
 */
void checkFramebuffer(GLuint type, GLint res) {
    if (g_Framebuffer[type] > 0)
    {
        GLenum const error = glGetError();
        GLenum const status = glCheckNamedFramebufferStatus(g_Framebuffer[type], GL_FRAMEBUFFER);

        if ((status == GL_FRAMEBUFFER_COMPLETE) && (error == GL_NO_ERROR))
            res = g_Resolution;
        else
            printf("FBO creation failed, glCheckFramebufferStatus() = 0x%x, glGetError() = 0x%x\n", status, error);
    }
}

/**
 * @brief Initialization function.
 * Sets up resources, algorithms and textures.
 */
void init() {
    // Default scene distance
    Variables::Transform.SceneZOffset = 24.0f;

    // Set OpenGL state variables
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    // Load a create texture for scene
    g_Textures[Diffuse] = Tools::Texture::CreateFromFile(TEXTURE_FILE_NAME);

    // Setup algorithms
    createVirtualFramebuffers();
    g_ClassicShadowMap.setup();
    g_RTWBackwards.setup();
    g_RTWForward.setup();

}


// Include GUI and control stuff
#include "controls.hpp"
